import { AngularFirestoreCollection, AngularFirestore, DocumentChangeAction, QueryFn, QueryGroupFn } from '@angular/fire/firestore';
import { EntityStore, EntityState, ActiveState, getEntityType } from '@datorama/akita';
import firebase from 'firebase/app';
import { WriteOptions, SyncOptions, PathParams, UpdateCallback, AtomicWrite } from '../utils/types';
import { Observable } from 'rxjs';
export declare type CollectionState<E = any> = EntityState<E, string> & ActiveState<string>;
export declare type DocOptions = {
    path: string;
} | {
    id: string;
};
export declare type GetRefs<idOrQuery> = idOrQuery extends (infer I)[] ? firebase.firestore.DocumentReference[] : idOrQuery extends string ? firebase.firestore.DocumentReference : firebase.firestore.CollectionReference;
/** check is an Atomic write is a transaction */
export declare function isTransaction(write: AtomicWrite): write is firebase.firestore.Transaction;
export declare class CollectionService<S extends EntityState<EntityType, string>, EntityType = getEntityType<S>> {
    protected store?: EntityStore<S>;
    private collectionPath?;
    private idsToListen;
    private memo;
    protected db: AngularFirestore;
    protected onCreate?(entity: EntityType, options: WriteOptions): any;
    protected onUpdate?(entity: Partial<EntityType>, options: WriteOptions): any;
    protected onDelete?(id: string, options: WriteOptions): any;
    /** If true, it will multicast observables from the same ID */
    protected useMemorization: boolean;
    constructor(store?: EntityStore<S>, collectionPath?: string, db?: AngularFirestore);
    private fromMemo;
    protected getPath(options: PathParams): string;
    get idKey(): string;
    /** The path to the collection in Firestore */
    get path(): string;
    /** A snapshot of the path */
    get currentPath(): string;
    get resetOnUpdate(): boolean;
    get mergeRef(): boolean;
    /**
     * The Angular Fire collection
     * @notice If path is an observable, it becomes an observable.
     */
    get collection(): AngularFirestoreCollection<EntityType>;
    /**
     * Function triggered when adding/updating data to firestore
     * @note should be overridden
     */
    formatToFirestore(entity: Partial<EntityType>): any;
    /**
     * Function triggered when getting data from firestore
     * @note should be overridden
     */
    formatFromFirestore(entity: any): EntityType;
    /**
     * Sync a store with a collection query of Firestore
     * @param path Path of the collection in Firestore
     * @param queryFn A query function to filter document of the collection
     * @param syncOptions Options about the store to sync with Firestore
     * @example
     * service.syncCollection({ storeName: 'movies-latest', loading: false }).subscribe();
     */
    syncCollection(syncOptions?: Partial<SyncOptions>): Observable<DocumentChangeAction<EntityType>[]>;
    /**
     * @example
     * service.syncCollection(activePath$).subscribe();
     */
    syncCollection(path: string, syncOptions?: Partial<SyncOptions>): Observable<DocumentChangeAction<EntityType>[]>;
    /**
     * @example
     * service.syncCollection(ref => ref.limit(10), { storeName: 'movie-latest'}).subscribe();
     */
    syncCollection(query: QueryFn, syncOptions?: Partial<SyncOptions>): Observable<DocumentChangeAction<EntityType>[]>;
    /**
     * @example
     * service.syncCollection('movies', ref => ref.limit(10), { loading: false }).subscribe();
     */
    syncCollection(path: string, queryFn?: QueryFn, syncOptions?: Partial<SyncOptions>): Observable<DocumentChangeAction<EntityType>[]>;
    /**
     * Sync a store with a collection group
     * @param collectionId An id of
     * @param queryFn A query function to filter document of the collection
     * @param syncOptions Options about the store to sync with Firestore
     */
    syncCollectionGroup(syncOptions?: Partial<SyncOptions>): Observable<DocumentChangeAction<EntityType>[]>;
    syncCollectionGroup(queryGroupFn?: QueryGroupFn): Observable<DocumentChangeAction<EntityType>[]>;
    syncCollectionGroup(collectionId: string, syncOptions?: Partial<SyncOptions>): Observable<DocumentChangeAction<EntityType>[]>;
    syncCollectionGroup(collectionId: string, queryGroupFn?: QueryGroupFn, syncOptions?: Partial<SyncOptions>): Observable<DocumentChangeAction<EntityType>[]>;
    /**
     * Sync the store with several documents
     * @param ids An array of ids or an observable of ids
     * @param syncOptions Options on the store to sync to
     */
    syncManyDocs(ids: string[], syncOptions?: Partial<SyncOptions>): any;
    /** @deprecated Use syncManyDocs inside a switchMap instead of passing an observable */
    syncManyDocs(ids$: Observable<string[]>, syncOptions?: Partial<SyncOptions>): any;
    /**
     * Stay in sync with one document
     * @param docOptions An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     * @param syncOptions Options on the store to sync to
     */
    syncDoc(docOptions: DocOptions, syncOptions?: Partial<SyncOptions>): Observable<EntityType>;
    /**
     * Stay in sync with the active entity and set it active in the store
     * @param options A list of ids or An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     * @param syncOptions Options on the store to sync to
     */
    syncActive(options: S['active'] extends any[] ? string[] : DocOptions, syncOptions?: Partial<SyncOptions>): S['active'] extends any[] ? Observable<EntityType[]> : Observable<EntityType>;
    /** Return the reference of the document(s) or collection */
    getRef(options?: Partial<SyncOptions>): firebase.firestore.CollectionReference;
    getRef(ids?: string[], options?: Partial<SyncOptions>): firebase.firestore.DocumentReference[];
    getRef(id?: string, options?: Partial<SyncOptions>): firebase.firestore.DocumentReference;
    /** Return the current value of the path from Firestore */
    getValue(options?: Partial<SyncOptions>): Promise<EntityType[]>;
    getValue(ids?: string[], options?: Partial<SyncOptions>): Promise<EntityType[]>;
    getValue(query?: QueryFn, options?: Partial<SyncOptions>): Promise<EntityType[]>;
    getValue(id: string, options?: Partial<SyncOptions>): Promise<EntityType>;
    /** Listen to the change of values of the path from Firestore */
    valueChanges(options?: Partial<PathParams>): Observable<EntityType[]>;
    valueChanges(ids?: string[], options?: Partial<PathParams>): Observable<EntityType[]>;
    valueChanges(query?: QueryFn, options?: Partial<PathParams>): Observable<EntityType[]>;
    valueChanges(id: string, options?: Partial<PathParams>): Observable<EntityType>;
    /**
     * Create a batch object.
     * @note alias for `angularFirestore.firestore.batch()`
     */
    batch(): firebase.firestore.WriteBatch;
    /**
     * Run a transaction
     * @note alias for `angularFirestore.firestore.runTransaction()`
     */
    runTransaction(cb: Parameters<firebase.firestore.Firestore['runTransaction']>[0]): Promise<unknown>;
    /**
     * Create or update documents
     * @param documents One or many documents
     * @param options options to write the document on firestore
     */
    upsert<D extends (Partial<EntityType> | Partial<EntityType>[])>(documents: D, options?: WriteOptions): Promise<D extends (infer I)[] ? string[] : string>;
    /**
     * Add a document or a list of document to Firestore
     * @param docs A document or a list of document
     * @param options options to write the document on firestore
     */
    add<D extends (Partial<EntityType> | Partial<EntityType>[])>(documents: D, options?: WriteOptions): Promise<D extends (infer I)[] ? string[] : string>;
    /**
     * Remove one or several document from Firestore
     * @param id A unique or list of id representing the document
     * @param options options to write the document on firestore
     */
    remove(id: string | string[], options?: WriteOptions): Promise<void>;
    /** Remove all document of the collection */
    removeAll(options?: WriteOptions): Promise<void>;
    /**
     * Update one or several document in Firestore
     */
    update(entity: Partial<EntityType> | Partial<EntityType>[], options?: WriteOptions): Promise<void>;
    update(id: string | string[], entityChanges: Partial<EntityType>, options?: WriteOptions): Promise<void>;
    update(ids: string | string[], stateFunction: UpdateCallback<EntityType>, options?: WriteOptions): Promise<firebase.firestore.Transaction[]>;
}
